C
C     Include modlue with type definitions
C
      INCLUDE 'inc_types.f'
C
C     Start of TriTok CFD code
C
      PROGRAM Tritok
C
C      
C      
C           3D BEM based fluid flow solver
C       
C           written by 
C
C           - Jure Ravnik, jure.ravnik@uni-mb.si
C           - Leopold Skerget
C           - Zoran Zunic
C
C           University of Maribor
C           Faculty of Mechanical Engineering
C           Smetanova 17, SI-2000, Maribor, Slovenia
C
C     kode = 0  - resuje kinematiko za robne vrtincnosti, predpise hitrost iz BIC
C     kode = 1  - NE resuje kinematike za robne vrtincnosti, predpise hitrost iz BIC
C     kode = 2  - NE resuje kinematike za robne vrtincnosti, open B.C. za hitrost
C     kode = 3  - resuje kinematiko za robne vrtincnosti, open B.C. za hitrost
C
C     EQNS iBw iDv iDw iDT iDC
C
C     iBw = 1  - single domain resitev enacbe kinematike za robne vrtincnosti
C     iBw = 2  - single domain resitev enacbe kinematike za robne vrtincnosti
C                z linearno aproksimacijo za obmocni clen (DELA SLABO)
C     iBw = 3  - single domain resitev enacbe kinematike za robne vrtincnosti
C                s pomocjo rotacijskih matrik
C     iDv = 1  - least squares resitev predolocenega sistema enacb za subdomain kinematiko
C                za obmocne hitrosti
C     iDv = 2  - resitev dolocenega sistema enacb za subdomain kinematiko
C                za obmocne hitrosti
C     iDw = 1  - least squares resitev predolocenega sistema enacb za subdomain kinetiko
C                za obmocne vrtincnosti
C     iDw = 2  - resitev dolocenega sistema enacb za subdomain kinetiko za obmocne vrtincnosti
C     iDw = 3  - least squares resitev predolocenega sistema enacb za subdomain kinetiko
C                za obmocne vrtincnosti kjer sta viskoznost in beta spremenljivi
C     iDT = 1  - least squares resitev predolocenega sistema enacb za subdomain kinetiko
C                za obmocne temperature
C     iDT = 2  - resitev dolocenega sistema enacb za subdomain kinetiko za obmocne temperature
C     iDT = 3  - resitev predolocenega sistema enacb za subdomain kinetiko za obmocne temperature
C                z spremeljivimi snovnimi lastnostmi
C     iDC = 1  - moj (laplace) least squares resitev predolocenega sistema enacb za subdomain
C                kinetiko za DC enacbo s spremenljivo difuzivnostjo
C     iDC = 2  - parametrix least squares resitev predolocenega sistema enacb za subdomain
C                kinetiko za DC enacbo s spremenljivo difuzivnostjo
C     iDC = 3  - parametrix resitev dolocenega sistema enacb za subdomain
C                kinetiko za DC enacbo s spremenljivo difuzivnostjo
C     iDC = 4  - parametrix least squares resitev predolocenega sistema enacb za subdomain
C                kinetiko za DC enacbo s spremenljivo difuzivnostjo, hitrost je v sistemski,
C                rabi samo eno itreracijo v linearnih primerih.
C
C
      USE inc_types

      TYPE(meshType)  :: mesh
      TYPE(gausstype) :: gauss
      TYPE(IOtype)    :: io
      TYPE(inputtype) :: inp      
      TYPE(KMpointer) :: kmp
      TYPE(CPUtype)   :: cpu
      TYPE(countType) :: cnt    
      TYPE(penv) :: env
      TYPE(particleType), ALLOCATABLE :: part(:) ! seznam delcev
      
      TYPE(matrix) :: VTEsysmX,VTErhsmX,VTEprecX ! vorticity transport eqaution sysm in rhsm X
      TYPE(matrix) :: VTEsysmY,VTErhsmY,VTEprecY
      TYPE(matrix) :: VTEsysmZ,VTErhsmZ,VTEprecZ
      TYPE(matrix) :: Tsysm,Trhsm,Tprec      ! energy equation
      TYPE(matrix) :: DCsysm,DCrhsm,DCprec      ! general diff-conv. equation
      TYPE(matrix) :: KMsysm,KMrhsm,KMprec     ! kinematics - for domain velocities
C
C     Macro integrals matrices
C
      REAL(8), ALLOCATABLE :: smatH(:,:), smatG(:,:)
      REAL(8), ALLOCATABLE :: smatB(:,:) ! source term (and FD time term)
      REAL(8), ALLOCATABLE :: smatAbdX(:,:),smatAbdY(:,:),smatAbdZ(:,:) ! advection boundary-domain
      REAL(8), ALLOCATABLE :: smatHtx(:,:),smatHty(:,:),smatHtz(:,:) ! kinematics, Ht
      REAL(8), ALLOCATABLE :: smatDx(:,:),smatDy(:,:),smatDz(:,:) ! kinematics, D
C
C     Single domain kinematics matrices
C
      REAL(8), ALLOCATABLE :: sdH(:,:)
      REAL(8), ALLOCATABLE :: sdHtx(:,:),sdHty(:,:),sdHtz(:,:) ! kinematics, Ht
      REAL(8), ALLOCATABLE :: sdDx(:,:),sdDy(:,:),sdDz(:,:) ! kinematics, D
      REAL(8), ALLOCATABLE :: sdDxB(:,:),sdDyB(:,:),sdDzB(:,:) ! kinematics, D
      REAL(8), ALLOCATABLE :: sdAx(:,:), sdPivotx(:) ! system matrix
      REAL(8), ALLOCATABLE :: sdAy(:,:), sdPivoty(:) ! system matrix
      REAL(8), ALLOCATABLE :: sdAz(:,:), sdPivotz(:) ! system matrix            

      REAL(8), ALLOCATABLE :: KMprecv(:) ! preconditioning vectors
      REAL(8), ALLOCATABLE :: VTEprecvX(:),VTEprecvY(:),VTEprecvZ(:) ! preconditioning vectors      
      REAL(8), ALLOCATABLE :: Tprecv(:), DCprecv(:)      
C
C     Field functions
C     
      REAL(8), ALLOCATABLE :: vorticity(:,:)
      REAL(8), ALLOCATABLE :: qvorticity(:,:)
      REAL(8), ALLOCATABLE :: ptsvorticity(:,:)  ! previous time step
      REAL(8), ALLOCATABLE :: pptsvorticity(:,:) ! pred previous time step               
      REAL(8), ALLOCATABLE :: pnlsvorticity(:,:) ! previous nonlinear iteration   
      REAL(8), ALLOCATABLE :: rhsv(:,:)  ! vorticity right hand side vector 
      
      REAL(8), ALLOCATABLE :: velocity(:,:),pnlsvelocity(:,:)  

      REAL(8), ALLOCATABLE :: temp(:)
      REAL(8), ALLOCATABLE :: qtemp(:)
      REAL(8), ALLOCATABLE :: ptstemp(:)  ! previous time step
      REAL(8), ALLOCATABLE :: pptstemp(:) ! pred previous time step               
      REAL(8), ALLOCATABLE :: pnlstemp(:) ! previous nonlinear iteration   
      REAL(8), ALLOCATABLE :: Trhsv(:)  ! temperature right hand side vector
      REAL(8), ALLOCATABLE :: Ttime(:)  ! temperature source due to previous time step

      REAL(8), ALLOCATABLE :: dc(:)    ! polje za splosno D-C enacbo, s spremenljivo diff.
      REAL(8), ALLOCATABLE :: qdc(:)
      REAL(8), ALLOCATABLE :: ptsdc(:)  ! previous time step
      REAL(8), ALLOCATABLE :: pptsdc(:) ! pred previous time step               
      REAL(8), ALLOCATABLE :: pnlsdc(:) ! previous nonlinear iteration   
      REAL(8), ALLOCATABLE :: DCrhsv(:)  ! DC right hand side vector
      REAL(8), ALLOCATABLE :: DCtime(:)  ! DC time derivative right hand side vector
      REAL(8), ALLOCATABLE :: DCdiff(:)  ! DC difuzivnost
      REAL(8), ALLOCATABLE :: DCqDiff(:)  ! DC difuzivnost v macro tockah
      REAL(8), ALLOCATABLE :: DCdiffGrad(:,:)  ! gradient DC difuzivnost   
      REAL(8), ALLOCATABLE :: DCdiffLap(:)  ! laplace DC difuzivnost       

      REAL(8), ALLOCATABLE :: tmp(:)  ! zacasno, pomozno polje
C     nanofluids
      REAL(8) nanoA, nanoB, nanoC
      REAL(8), ALLOCATABLE :: PartVolFrac(:) ! particle volume fraction
c     ferrofluids
      REAL(8), ALLOCATABLE :: mfs(:,:)  ! magnetic field strength \vec H
      REAL(8), ALLOCATABLE :: hshg(:,:)  ! \vec H \cdot \vec H, H skalarno H
      REAL(8), ALLOCATABLE :: kbf(:,:)  ! Kelvin body force
      REAL(8), ALLOCATABLE :: tempGrad(:,:)  ! Temperature gradient
c     energy equation with variable material propersites
      REAL(8), ALLOCATABLE :: Density(:)  ! nondimensional density
      REAL(8), ALLOCATABLE :: HeatCapacity(:)  ! nondimensional heat capacity
      REAL(8), ALLOCATABLE :: ThermalConductivity(:)  ! nondimensional thermal conductivity
      REAL(8), ALLOCATABLE :: qThermalConductivity(:)  ! nondimensional thermal conductivity in flux points
      REAL(8), ALLOCATABLE :: gradTC(:,:)     ! gradient of nondimensional thermal conductivity
      REAL(8), ALLOCATABLE :: dRhoCpdt(:)  ! d(rho*cp)/dt
      REAL(8), ALLOCATABLE :: prhocp(:)    ! rho*cp in previous time step
c     vorticity transport equation with variable material propersites
      REAL(8), ALLOCATABLE :: Viscosity(:)  ! nondimensional mu=mu^dim / mu^inReynolds
      REAL(8), ALLOCATABLE :: qViscosity(:)  ! nondimensional Viscosity  in flux points
      REAL(8), ALLOCATABLE :: GradVisc(:,:)   ! gradient of nondimensional Viscosity
      REAL(8), ALLOCATABLE :: cvte(:)  ! nondimensional beta=beta^dim / beta^inRayleigh
      REAL(8), ALLOCATABLE :: DvaEpsGVisc(:,:) !  2\epika\cdot\nabla\mu
      REAL(8), ALLOCATABLE :: CuOmCrGrVi(:,:)  !  \nabla\time\omega\times\nabla\mu


      REAL cptime
      REAL rcpu

      INTEGER i,j,itim,istop
      INTEGER eqn,ierr
c
C____ Init measuring cpu time
c
      cpu%t0=cptime(0.)
      cpu%t00=cpu%t0
      CALL InitCpu(cpu,itim)
c 
C____ Version data 
c
      CALL DatumInUra(inp%StartTime)
      inp%IDname='TriTok'
      inp%IDversion='7.2'
      inp%IDdate='januar 2018'
      inp%INTversion=1
      inp%RSTversion=4
c
C____ Init random generator
c
      CALL RANDOM_SEED
c
C____ MPI init
c
      CALL par_init(env)
c
C____ Input / output files
c
      CALL SetUpIO(env,io,inp)
c 
C____ Read input file
c
      CALL ReadInputFile(env,mesh,io,inp,gauss)
c
C____ Set up Gauss quadrature positions and weights
c
      CALL GaussPosWeights(gauss)
C
C     Prebere mrezo 
C
      CALL ReadMeshGid3D(env,io,inp,mesh,gauss)
      CALL VmesniCas(cpu,itim)
C
C     Generira tocke xq in povezave ibf za nezvezni del mreze
C
      CALL GenMakroMesh(mesh)         
      
      IF (env%myproc.EQ.1) THEN
      WRITE(io%l,'(A,I8)') "nnodes      =",mesh%nnodes  
      WRITE(io%l,'(A,I8)') "nq          =",mesh%nq
      WRITE(io%l,'(A,I8)') "vozlisc     =",mesh%nq+mesh%nnodes 
      WRITE(io%l,'(A)') ""
      WRITE(io%l,'(A,I8)') "nbnodes     =",mesh%nbnodes
      WRITE(io%l,'(A,I8)') "flux na robu=",mesh%nbelem*4
      WRITE(io%l,'(A,I8)') "robnih      =",mesh%nbelem*4+mesh%nbnodes
      WRITE(io%l,'(A)') ""
      END IF
      CALL VmesniCas(cpu,itim)
C 
C     Allocate field functions
C
      ALLOCATE (velocity(mesh%nnodes,3))
      ALLOCATE (vorticity(mesh%nnodes,3))
      ALLOCATE (qvorticity(mesh%nq,3))        
      ALLOCATE (pnlsvelocity(mesh%nnodes,3)) ! previous non linear iteration
      ALLOCATE (ptsvorticity(mesh%nnodes,3))  ! previous time step
      ALLOCATE (pptsvorticity(mesh%nnodes,3)) ! pred previous time step               
      ALLOCATE (pnlsvorticity(mesh%nnodes,3)) ! previous nonlinear iteration
      ALLOCATE (rhsv(mesh%nnodes,3))   ! body force (includes time derivatives)          

      ALLOCATE (temp(mesh%nnodes))
      ALLOCATE (qtemp(mesh%nq))        
      ALLOCATE (ptstemp(mesh%nnodes))  ! previous time step
      ALLOCATE (pptstemp(mesh%nnodes)) ! pred previous time step               
      ALLOCATE (pnlstemp(mesh%nnodes)) ! previous nonlinear iteration
      ALLOCATE (Trhsv(mesh%nnodes))    ! body force (total)
      ALLOCATE (Ttime(mesh%nnodes))    ! body force (only time derivatives)

      ALLOCATE (dc(mesh%nnodes))
      ALLOCATE (qdc(mesh%nq))        
      ALLOCATE (ptsdc(mesh%nnodes))  ! previous time step
      ALLOCATE (pptsdc(mesh%nnodes)) ! pred previous time step               
      ALLOCATE (pnlsdc(mesh%nnodes)) ! previous nonlinear iteration
      ALLOCATE (DCrhsv(mesh%nnodes))    ! body force (time derivatives + sources)
      ALLOCATE (DCtime(mesh%nnodes))    ! body force (time derivatives)
      ALLOCATE (DCdiff(mesh%nnodes))    ! difuzivnost
      ALLOCATE (DCqDiff(mesh%nq))    ! difuzivnost v macro tockah
      ALLOCATE (DCdiffGrad(mesh%nnodes,3))    ! gradient difuzivnosti 
      ALLOCATE (DCdiffLap(mesh%nnodes))    ! nabla kvadrat difuzivnosti

      ALLOCATE (tmp(mesh%nnodes))    ! zacasno pomozno polje

C 
C     Prebere BiC datoteko
C
      CALL ReadBic(env,io,inp,mesh,velocity,vorticity,qvorticity,temp,qtemp,dc,qdc)
      IF (env%myproc.EQ.1) CALL OutputInitial(io,inp,mesh,velocity,vorticity,temp,dc)

C
C     V bic je koncna razdelitev vozlisc po stenah
C     za vsako robno vozlisce najdem najblizje vozlisce, ki
C     ni na isti steni. To potrebujem za dv/dn=0 robni pogoj
C
      CALL FillKobc(mesh)
      CALL VmesniCas(cpu,itim)
c
C____ Divide boundary nodes across processors
c                          
      CALL SetEnvGlob(env,mesh)
      CALL DivideNodes(env,mesh)
c
c     Estimate requred RAM per processor
c
      IF (env%myproc.EQ.1) CALL EstRAM(env,mesh,io,inp)
      IF (inp%iestr.EQ.1) THEN
        CALL WarnErr(env,io,inp,0,'Tritok','Estimated RAM only, exiting!',0)
        CALL StopProgram(env,io,inp,0)
      END IF
C
C     Nanofluid
C
      IF (inp%inano.EQ.1) THEN
        CALL NanofluidProp(inp,io,env,nanoA,nanoB,nanoC)
        IF (inp%iDT.EQ.3) nanoC=1.0D0 ! ker racunam z variable properties
      ELSE
        nanoA=1.0D0
        nanoB=1.0D0
        nanoC=1.0D0
      END IF
C
C     Ferrofluid
C
      ALLOCATE (kbf(mesh%nnodes,3)) ! Kelvin body force
      IF (inp%iFerro.GT.0) THEN
        WRITE (io%l,'(A)') "Simulation of ferrofluid"
        WRITE (io%l,'(A,G18.9)') "Magnetic Rayleigh number =",inp%Ram
        WRITE (io%l,'(A,G18.9)') "Beta * Delta T =",inp%bdt
        WRITE (io%l,'(A,G18.9/)') "hi_0 =",inp%hi0

        ALLOCATE (mfs(mesh%nnodes,3)) ! magnetic field strength \vec H
        ALLOCATE (hshg(mesh%nnodes,3)) ! grad \vec H \cdot \vec H, gradient (H skalarno H)
        ALLOCATE (tempGrad(mesh%nnodes,3)) ! Temperature gradient
        CALL SetUpMagneticField(mesh,inp,mfs,hshg)
      ELSE
        kbf=0.0D0
        inp%ram=0.0D0
      END IF
C
C     Single Domain kinematika
C

C     Robne matrike
      IF (inp%iBw.NE.0) THEN
        ALLOCATE (sdH(env%zac:env%kon,mesh%nbnodes))  ! diffusion
        ALLOCATE (sdHtx(env%zac:env%kon,mesh%nbnodes)) ! doktorat, enacba (4.7)
        ALLOCATE (sdHty(env%zac:env%kon,mesh%nbnodes)) ! to je za prvi integral
        ALLOCATE (sdHtz(env%zac:env%kon,mesh%nbnodes)) ! na desni

C     Kvadratna interpolacija vseh funkcij
      IF (inp%iBw.EQ.1) THEN
C       Sistemske matrike in predpogojevalci
        ALLOCATE (sdAx(env%zac:env%kon,mesh%nbnodes))
        ALLOCATE (sdAy(env%zac:env%kon,mesh%nbnodes))
        ALLOCATE (sdAz(env%zac:env%kon,mesh%nbnodes))
        ALLOCATE (sdPivotx(env%zac:env%kon))
        ALLOCATE (sdPivoty(env%zac:env%kon))
        ALLOCATE (sdPivotz(env%zac:env%kon))
C       Obmocne matrike
        ALLOCATE (sdDx(env%zac:env%kon,mesh%nnodes)) ! doktorat, enacba (4.7)
        ALLOCATE (sdDy(env%zac:env%kon,mesh%nnodes)) ! to je za drugi integral
        ALLOCATE (sdDz(env%zac:env%kon,mesh%nnodes)) ! na desni
C
C       Compute or read single domain kinematics integrals
C
        CALL sdKm_corInt(env,io,inp,mesh,gauss,sdH,sdHtx,sdHty,sdHtz,sdDx,sdDy,sdDz)
C
C       Form single domain kinematics system matrices and preconditioners
C
        CALL sdKm_sysm(1,env,mesh,sdAx,sdDx,sdDy,sdDz)
        CALL sdKm_sysm(2,env,mesh,sdAy,sdDx,sdDy,sdDz)
        CALL sdKm_sysm(3,env,mesh,sdAz,sdDx,sdDy,sdDz)
c        CALL CalculateRotationMatrices(mesh) ! test

C     Omega na desni strani linearno interpolirana
      ELSE IF (inp%iBw.EQ.2) THEN ! omega interpolirana linearno
C       Sistemske matrike in predpogojevalci
        ALLOCATE (sdAx(env%zac:env%kon,mesh%nbnodes))
        ALLOCATE (sdAy(env%zac:env%kon,mesh%nbnodes))
        ALLOCATE (sdAz(env%zac:env%kon,mesh%nbnodes))
        ALLOCATE (sdPivotx(env%zac:env%kon))
        ALLOCATE (sdPivoty(env%zac:env%kon))
        ALLOCATE (sdPivotz(env%zac:env%kon))
C       Obmocni integrali izracuznani z linearno interpolacijo za omego
        ALLOCATE (sdDx(env%zac:env%kon,mesh%nnodes8)) ! doktorat, enacba (4.7)
        ALLOCATE (sdDy(env%zac:env%kon,mesh%nnodes8)) ! to je za drugi integral
        ALLOCATE (sdDz(env%zac:env%kon,mesh%nnodes8)) ! na desni
c       robni del (kvadratna interpolacija)
        ALLOCATE (sdDxB(env%zac:env%kon,mesh%nbnodes)) ! doktorat, enacba (4.7)
        ALLOCATE (sdDyB(env%zac:env%kon,mesh%nbnodes)) ! to je za drugi integral
        ALLOCATE (sdDzB(env%zac:env%kon,mesh%nbnodes)) ! na desni

        CALL sdKm_corInt8(env,io,inp,mesh,gauss,sdH,sdHtx,sdHty,sdHtz,sdDx,sdDy,sdDz,sdDxB,sdDyB,sdDzB)
C
C       Form single domain kinematics system matrices
C
        CALL sdKm_sysm8(1,env,mesh,sdAx,sdDxB,sdDyB,sdDzB)
        CALL sdKm_sysm8(2,env,mesh,sdAy,sdDxB,sdDyB,sdDzB)
        CALL sdKm_sysm8(3,env,mesh,sdAz,sdDxB,sdDyB,sdDzB)

C     We recognise that normal vorticity component is zero at the wall,
C     so we solve only for tangential components
      ELSE IF (inp%iBw.EQ.3) THEN
C
C       For each source point, we find the largest component of the normal
C       This vorticity component, will be calculated by use of known w_normal
C
        CALL AnalyseSourcePointNormals(mesh)

C       Sistemske matrike in predpogojevalci
        ALLOCATE (sdAx(env%zac:env%kon,mesh%nbnodes))
        ALLOCATE (sdAy(env%zac:env%kon,mesh%nbnodes))
        ALLOCATE (sdAz(env%zac:env%kon,mesh%nbnodes))
        ALLOCATE (sdPivotx(env%zac:env%kon))
        ALLOCATE (sdPivoty(env%zac:env%kon))
        ALLOCATE (sdPivotz(env%zac:env%kon))
C       Obmocne matrike
        ALLOCATE (sdDx(env%zac:env%kon,mesh%nnodes)) ! doktorat, enacba (4.7)
        ALLOCATE (sdDy(env%zac:env%kon,mesh%nnodes)) ! to je za drugi integral
        ALLOCATE (sdDz(env%zac:env%kon,mesh%nnodes)) ! na desni
C
C       Compute or read single domain kinematics integrals
C
        CALL sdKm_corInt(env,io,inp,mesh,gauss,sdH,sdHtx,sdHty,sdHtz,sdDx,sdDy,sdDz)
C
C       Form single domain kinematics system matrices and preconditioners
C
        CALL sdKm_sysmBCN(1,env,mesh,sdAx,sdDx,sdDy,sdDz)
        CALL sdKm_sysmBCN(2,env,mesh,sdAy,sdDx,sdDy,sdDz)
        CALL sdKm_sysmBCN(3,env,mesh,sdAz,sdDx,sdDy,sdDz)

      END IF


      CALL VmesniCas(cpu,itim)        

C
C     Preconditioners
C

        IF (env%nproc.EQ.1) THEN
c         1 proc:  0 2 2 500 5 1.E-6
          CALL FormPREfm(2,mesh%nbnodes,sdPivotx,sdAx,rcpu,ierr)  ! pret=2 - LU faktorizacija
          CALL FormPREfm(2,mesh%nbnodes,sdPivoty,sdAy,rcpu,ierr)  ! pret=2 - LU faktorizacija
          CALL FormPREfm(2,mesh%nbnodes,sdPivotz,sdAz,rcpu,ierr)  ! pret=2 - LU faktorizacija
        ELSE
c          vec proc:  2 1 2 500 5 1.E-6
          CALL pFormPREfm(env,1,mesh%nbnodes,env%nmn,sdPivotx,sdAx,rcpu,ierr) ! pret=1 - diagonal
          CALL pFormPREfm(env,1,mesh%nbnodes,env%nmn,sdPivoty,sdAy,rcpu,ierr) ! pret=1 - diagonal
          CALL pFormPREfm(env,1,mesh%nbnodes,env%nmn,sdPivotz,sdAz,rcpu,ierr) ! pret=1 - diagonal
        END IF

      IF (env%myproc.EQ.1) THEN              
        WRITE (io%l,'(A)') "single domain kinematika za robne vrtincnosti"
        WRITE (io%l,'(A,I7)')  "stevilo enacb   =",mesh%nbnodes
        WRITE (io%l,'(A,I7/)') "stevilo neznank =",mesh%nbnodes
        IF (inp%iBw.EQ.2) THEN
          WRITE (io%l,'(A)') "Uporabljam linearno aproksimacijo za vrtincnost na desni!"
          WRITE (io%l,'(A,I7/)')  "stevilo linearnih vozlisc   =",mesh%nnodes8
        END IF
      END IF
      CALL VmesniCas(cpu,itim)

      END IF
C
C     Macro
C

C
C     Determine kinematics knowns and unknowns
C
      CALL DetKMunk3(mesh,kmp)
       
C
C     Integracija, shranim v pravokotne matrike
C      
      ALLOCATE (smatH(mesh%nicnsp,mesh%npoc),smatG(mesh%nicnsp,mesh%npofc))
C
C     Kinetics matrices
C      
      ALLOCATE (smatB(mesh%nicnsp,mesh%npoc))   ! source term
      ALLOCATE (smatAbdx(mesh%nicnsp,mesh%npoc)) ! advection - boundary  part X
      ALLOCATE (smatAbdy(mesh%nicnsp,mesh%npoc)) ! advection - boundary  part Y
      ALLOCATE (smatAbdz(mesh%nicnsp,mesh%npoc)) ! advection - boundary  part Z
C
C     kinematics matrices
C            
      ALLOCATE (smatHtx(mesh%nicnpoc,mesh%npoc)) ! doktorat, enacba (4.7)
      ALLOCATE (smatHty(mesh%nicnpoc,mesh%npoc)) ! to je za prvi integral 
      ALLOCATE (smatHtz(mesh%nicnpoc,mesh%npoc)) ! na desni
      
      ALLOCATE (smatDx(mesh%nicnsp,mesh%npoc)) ! doktorat, enacba (4.7)
      ALLOCATE (smatDy(mesh%nicnsp,mesh%npoc)) ! to je za drugi integral 
      ALLOCATE (smatDz(mesh%nicnsp,mesh%npoc)) ! na desni            
C
C     Integration
C
      CALL corMin_kmdcSteady(env,io,inp,mesh,gauss,smatH,smatG,smatB,
     &                          smatAbdx,smatAbdy,smatAbdz,
     &                          smatHtx,smatHty,smatHtz,smatDx,smatDy,smatDz)          
      CALL VmesniCas(cpu,itim)
C
C     time scheme
C
      inp%beta =+1.50D00/inp%tstep
      inp%beta2=+2.00D00/inp%tstep
      inp%beta3=+0.50D00/inp%tstep
C
C     tvorim CRS sysm in rhsm za kinetiko (notranje vrtincnosti)
C
      IF (inp%iDw.EQ.1) THEN  ! predolocen sistem enacb
        CALL sMat2crsSysRhsB_count(1,mesh,VTEsysmX,VTErhsmX)
        CALL sMat2crsSysRhsB_count(2,mesh,VTEsysmY,VTErhsmY)
        CALL sMat2crsSysRhsB_count(3,mesh,VTEsysmZ,VTErhsmZ)
C       Set up preconditioning vector (for diagonal preconditioning)
        ALLOCATE(VTEprecvX(mesh%nunk(1)))
        ALLOCATE(VTEprecvY(mesh%nunk(2)))
        ALLOCATE(VTEprecvZ(mesh%nunk(3)))            

        CALL sMat2crsSysRhsB_w_fill(1,mesh,smatH,smatG,smatB,inp%beta,inp%ren/nanoA,VTEsysmX,VTErhsmX,
     &                                velocity,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz)
        CALL sMat2crsSysRhsB_w_fill(2,mesh,smatH,smatG,smatB,inp%beta,inp%ren/nanoA,VTEsysmY,VTErhsmY,
     &                                velocity,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz)
        CALL sMat2crsSysRhsB_w_fill(3,mesh,smatH,smatG,smatB,inp%beta,inp%ren/nanoA,VTEsysmZ,VTErhsmZ,
     &                                velocity,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz)

c       calculate preconditioner        
        CALL prelsqr2(VTEsysmX%neq,mesh%nunk(1),VTEsysmX%nnz,VTEprecvX,VTEsysmX%i,VTEsysmX%j,VTEsysmX%v)      
        CALL prelsqr2(VTEsysmY%neq,mesh%nunk(2),VTEsysmY%nnz,VTEprecvY,VTEsysmY%i,VTEsysmY%j,VTEsysmY%v)      
        CALL prelsqr2(VTEsysmZ%neq,mesh%nunk(3),VTEsysmZ%nnz,VTEprecvZ,VTEsysmZ%i,VTEsysmZ%j,VTEsysmZ%v)



      ELSE IF (inp%iDw.EQ.2) THEN  ! dolocen sistem enacb
        CALL sMat2crsSysRhsB_SQsetup(1,mesh)
        CALL sMat2crsSysRhsB_SQsetup(2,mesh)
        CALL sMat2crsSysRhsB_SQsetup(3,mesh)

        CALL sMat2crsSysRhsB_SQcount(1,mesh,VTEsysmX,VTErhsmX)
        CALL sMat2crsSysRhsB_SQcount(2,mesh,VTEsysmY,VTErhsmY)
        CALL sMat2crsSysRhsB_SQcount(3,mesh,VTEsysmZ,VTErhsmZ)

        CALL sMat2crsSysRhsB_w_SQfill(1,mesh,smatH,smatG,smatB,inp%beta,inp%ren/nanoA,VTEsysmX,VTErhsmX,
     &                                velocity,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz)
        CALL sMat2crsSysRhsB_w_SQfill(2,mesh,smatH,smatG,smatB,inp%beta,inp%ren/nanoA,VTEsysmY,VTErhsmY,
     &                                velocity,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz)
        CALL sMat2crsSysRhsB_w_SQfill(3,mesh,smatH,smatG,smatB,inp%beta,inp%ren/nanoA,VTEsysmZ,VTErhsmZ,
     &                                velocity,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz)

C       Set up preconditioning matrix : 2 - ilu, 1 - diag, 0 - none
        CALL CopyMatrix(VTEsysmX,VTEprecX)
        CALL FormPRMjr(inp%sqrs_prec,VTEprecX,VTEsysmX)

        CALL CopyMatrix(VTEsysmY,VTEprecY)
        CALL FormPRMjr(inp%sqrs_prec,VTEprecY,VTEsysmY)

        CALL CopyMatrix(VTEsysmZ,VTEprecZ)
        CALL FormPRMjr(inp%sqrs_prec,VTEprecZ,VTEsysmZ)

      ELSE IF (inp%iDw.EQ.3) THEN  ! variable viscosity and beta
C       Set up systems of equations
        CALL sMat2crsSysRhsB_count(1,mesh,VTEsysmX,VTErhsmX)
        CALL sMat2crsSysRhsB_count(2,mesh,VTEsysmY,VTErhsmY)
        CALL sMat2crsSysRhsB_count(3,mesh,VTEsysmZ,VTErhsmZ)
C       Set up preconditioning vector (for diagonal preconditioning)
        ALLOCATE(VTEprecvX(mesh%nunk(1)))
        ALLOCATE(VTEprecvY(mesh%nunk(2)))
        ALLOCATE(VTEprecvZ(mesh%nunk(3)))
C       Set up variable material properties
        ALLOCATE (Viscosity(mesh%nnodes))  ! nondimensional mu=mu^dim / mu^inReynolds
        ALLOCATE (qViscosity(mesh%nq))  ! nondimensional mu=mu^dim / mu^inReynolds (v flux tockah)
        ALLOCATE (GradVisc(mesh%nnodes,3)) ! gradient of viscosity
        ALLOCATE (cvte(mesh%nnodes)) ! nondimensional beta=beta^dim / beta^inRayleigh
        ALLOCATE( DvaEpsGVisc(mesh%nnodes,3)) !  2\epika\cdot\nabla\mu
        ALLOCATE( CuOmCrGrVi(mesh%nnodes,3) ) !  \nabla\time\omega\times\nabla\mu

        CALL SetUpViscCvte(env,io,inp,mesh,viscosity,cvte,temp)

      ELSE IF (inp%iDw.NE.0) THEN
        CALL WarnErr(env,io,inp,4,"Tritok","Wrong iDw code!",inp%iDw)
      END IF 
      IF (inp%iDw.GT.0) THEN
        IF (env%myproc.EQ.1) THEN
          WRITE (io%l,'(A)') "makro kinetika za notranje vrtincnosti"
          WRITE (io%l,'(A,I7)')  "stevilo enacb   X =",VTEsysmX%neq
          WRITE (io%l,'(A,I7)') "stevilo neznank X =",mesh%nunk(1)
          WRITE (io%l,'(A,I7)')  "stevilo enacb   Y =",VTEsysmY%neq
          WRITE (io%l,'(A,I7)') "stevilo neznank Y =",mesh%nunk(2)
          WRITE (io%l,'(A,I7)')  "stevilo enacb   Z =",VTEsysmZ%neq
          WRITE (io%l,'(A,I7/)') "stevilo neznank Z =",mesh%nunk(3)
        END IF
      END IF
C
C     Set up particles
C
      IF (inp%iPart.EQ.1) THEN
        CALL CoR_NeighbourList(env,mesh,io,inp)
        CALL CalNodeVolume(mesh)
        ALLOCATE (part(inp%npart))
        CALL SetUpParticles(inp,part,mesh)
        ALLOCATE (PartVolFrac(mesh%nnodes))
        CALL CalParticleVolumeFraction(env,io,inp,gauss,part,mesh,PartVolFrac)
        IF (inp%iBrown.EQ.1) CALL SetUpBrown(env,io,inp)
        IF (inp%iThph.EQ.1)  CALL SetUpThermophoresis(env,io,inp)
      END IF


C
C     tvorim CRS sysm in rhsm za kinetiko (notranje temperature)
C
      IF (inp%iDT.EQ.1) THEN ! predolocen sistem enacb
        CALL sMat2crsSysRhsB_count(4,mesh,Tsysm,Trhsm)
C       Set up preconditioning vector (for diagonal preconditioning)
        ALLOCATE(Tprecv(mesh%Tnunk))
        CALL sMat2crsSysRhsB_Temp_fill(mesh,smatH,smatG,smatB,inp%beta,inp%ren*inp%prn/nanoC,Tsysm,Trhsm,
     &                      velocity,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz)              
c       preconditioner            
        CALL prelsqr2(Tsysm%neq,mesh%Tnunk,Tsysm%nnz,Tprecv,Tsysm%i,Tsysm%j,Tsysm%v)
      ELSE IF (inp%iDT.EQ.2) THEN ! dolocen sistem enacb
        CALL sMat2crsSysRhsB_SQsetup(4,mesh)
        CALL sMat2crsSysRhsB_SQcount(4,mesh,Tsysm,Trhsm)
        CALL sMat2crsSysRhsB_Temp_SQfill(mesh,smatH,smatG,smatB,inp%beta,inp%ren*inp%prn/nanoC,Tsysm,Trhsm,
     &                      velocity,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz)

C       Set up preconditioning matrix : 2 - ilu, 1 - diag, 0 - none
        CALL CopyMatrix(Tsysm,Tprec)
        CALL FormPRMjr(inp%sqrs_prec,Tprec,Tsysm)
      ELSE IF (inp%iDT.EQ.3) THEN ! spremenljive lastnosti tekocine
C       Set up variable material properties
        ALLOCATE (Density(mesh%nnodes))  ! nondimensional density
        ALLOCATE (HeatCapacity(mesh%nnodes))  ! nondimensional heat capacity
        ALLOCATE (ThermalConductivity(mesh%nnodes))  ! nondimensional thermal conductivity
        ALLOCATE (qThermalConductivity(mesh%nq))  ! nondimensional thermal conductivity
        ALLOCATE (gradTC(mesh%nnodes,3))
        ALLOCATE (dRhoCpdt(mesh%nnodes)) ! d(rho*cp)/dt
        ALLOCATE (prhocp(mesh%nnodes))   ! rho*cp in previous time step

C       Set up  initial conditions
        CALL SetUpRhoCpLambdaAndBC(env,io,inp,mesh,gauss,Density,HeatCapacity,
     &                             ThermalConductivity,Temp,qTemp,velocity,inp%eeqm,Trhsv,cnt%rtime,1,PartVolFrac)

C       Count and set up crs matrices
        CALL sMat2crsSysRhsB_count(4,mesh,Tsysm,Trhsm)
C       Set up preconditioning vector (for diagonal preconditioning)
        ALLOCATE(Tprecv(mesh%Tnunk))

      ELSE IF (inp%iDT.NE.0) THEN
        CALL WarnErr(env,io,inp,4,"Tritok","Wrong iDT code!",inp%iDT)
      END IF
      IF (inp%iDT.GT.0) THEN
        IF (env%myproc.EQ.1) THEN
          WRITE (io%l,'(A)') "makro kinetika za notranje temperature"
          WRITE (io%l,'(A,I7)') "stevilo enacb   =",Tsysm%neq
          WRITE (io%l,'(A,I7/)') "stevilo neznank =",mesh%Tnunk
        END IF
      END IF
C
C     tvorim CRS sysm in rhsm za kinetiko (general diff-conv equation)
C
      IF (inp%iDC.EQ.1.OR.inp%iDC.EQ.2.OR.inp%iDC.EQ.4) THEN ! predolocen sistem enacb
        CALL sMat2crsSysRhsB_count(5,mesh,DCsysm,DCrhsm)
C       Set up preconditioning vector (for diagonal preconditioning)
        ALLOCATE(DCprecv(mesh%DCnunk))
      ELSE IF (inp%iDC.EQ.3) THEN ! dolocen sistem enacb
        CALL sMat2crsSysRhsB_SQsetup(5,mesh)
        CALL sMat2crsSysRhsB_SQcount(5,mesh,DCsysm,DCrhsm)
        CALL CopyMatrix(DCsysm,DCprec)
      ELSE IF (inp%iDC.NE.0) THEN
        CALL WarnErr(env,io,inp,4,"Tritok","Wrong iDC code!",inp%iDC)
      END IF

      IF (inp%iDC.GT.0) THEN
        IF (env%myproc.EQ.1) THEN
          WRITE (io%l,'(A)') "makro kinetika za splosno diff-conv enacbo"
          WRITE (io%l,'(A,I7)') "stevilo enacb   =",DCsysm%neq
          WRITE (io%l,'(A,I7/)') "stevilo neznank =",mesh%DCnunk
        END IF
      END IF

C
C     tvorim CRS sysm in rhsm za makro kinematiko za notranje hitrosti
C     ker so robni pogoji vedno predpisane hitrosti imam samo eno sistemsko matriko za vse tri enacbe
C     
      IF (inp%iDv.EQ.1) THEN
        CALL sMat2crsSysRhsKM3(mesh,smatH,KMsysm,KMrhsm,kmp)
C       Set up preconditioning vector (for diagonal preconditioning)
        ALLOCATE(KMprecv(kmp%nsmcol))
        CALL prelsqr2(KMsysm%neq,kmp%nsmcol,KMsysm%nnz,KMprecv,KMsysm%i,KMsysm%j,KMsysm%v)      
      ELSE IF (inp%iDv.EQ.2) THEN
        CALL sMat2crsSysRhsKM3sq(mesh,smatH,KMsysm,KMrhsm,kmp)
C       Set up preconditioning matrix : 2 - ilu, 1 - diag, 0 - none
        CALL CopyMatrix(KMsysm,KMprec)
        CALL FormPRMjr(inp%sqrs_prec,KMprec,KMsysm)
      ELSE IF (inp%iDv.NE.0) THEN
        CALL WarnErr(env,io,inp,4,"Tritok","Wrong iDv code!",inp%iDv)
      END IF
      IF (inp%iDv.GT.0) THEN
        IF (env%myproc.EQ.1) THEN
          WRITE (io%l,'(A)') "makro kinematika za notranje hitrosti"
          WRITE (io%l,'(A,I7)')  "stevilo enacb   =",KMsysm%neq
          WRITE (io%l,'(A,I7/)') "stevilo neznank =",kmp%nsmcol
        END IF
      END IF

c     Dont need this matrices anymore
      IF (inp%iDC.EQ.0.AND.inp%iDT.NE.3.AND.inp%iDw.NE.3) THEN
        DEALLOCATE (smatG)
        DEALLOCATE (smatH)
      END IF

      CALL VmesniCas(cpu,itim)
      
C      ----------------------------------------------------------------------------
C                                 PRE TIME LOOP INIT
C      ----------------------------------------------------------------------------      
      cnt%glnnlit=0
      cnt%nit=0
      cnt%tstep=0
      cnt%trtstep=0
      cnt%rtime=0.0D0
      pptsvorticity=vorticity ! u pred previous = u previous
      ptsvorticity=vorticity ! u previous time step = u      
      pnlsvorticity=vorticity
      pnlsvelocity=velocity

      cnt%ftime=1
      IF (inp%iDT.GT.0) THEN
        pptstemp=temp ! u pred previous = u previous
        ptstemp=temp ! u previous time step = u      
        pnlstemp=temp
        IF (inp%iDT.EQ.3) THEN
          DO i=1,mesh%nnodes
            prhocp(i)=Density(i)*HeatCapacity(i)
          END DO
        END IF
      END IF      

      IF (inp%iDC.GT.0) THEN
        pptsdc=dc ! u pred previous = u previous
        ptsdc=dc ! u previous time step = u      
        pnlsdc=dc
      END IF   
C
C     Read restart file
C          
      IF (inp%rst.EQ.1) CALL RrstF(env,mesh,io,inp,cnt,velocity,vorticity,
     &             ptsvorticity,pptsvorticity,qvorticity,pnlsvorticity,pnlsvelocity,
     &             temp,ptstemp,pptstemp,qtemp,pnlstemp,
     &             dc,ptsdc,pptsdc,qdc,pnlsdc)

C
C      ----------------------------------------------------------------------------
C                                 START TIME LOOP
C      ----------------------------------------------------------------------------
1618  CONTINUE
        cnt%trtstep=cnt%trtstep+1
        cnt%rtime=cnt%rtime+inp%tstep
        cnt%tstep=cnt%tstep+1
        cnt%tsnnlit=0  ! number of nonlinear iteration within time step=0
        
        inp%dlse=inp%lsqs_eps ! start with full solver accuracy
                  

        pptsvorticity=ptsvorticity ! u pred previous = u previous
        ptsvorticity=vorticity     ! u previous time step = u       
        DO j=1,3
          DO i=1,mesh%nnodes
c           od parcialnega odvoda po casu
            rhsv(i,j)=-ptsvorticity(i,j)*inp%beta2+pptsvorticity(i,j)*inp%beta3  
          END DO
        END DO

C       \p T / \p t  - temperature equation
        IF (inp%iDT.GT.0) THEN
          pptstemp=ptstemp ! u pred previous = u previous
          ptstemp=temp     ! u previous time step = u       
          DO i=1,mesh%nnodes
            Ttime(i)=-ptsTemp(i)*inp%beta2+pptsTemp(i)*inp%beta3
          END DO

          IF (inp%iDT.EQ.3) THEN
C           Calculate time derivative of rho*cp
            CALL TimeDerRhoCp(mesh,dRhoCpdt,prhocp,Density,HeatCapacity,inp%tstep)
          END IF

        END IF

C       general D-C equation
        IF (inp%iDC.GT.0) THEN         
          pptsdc=ptsdc ! u pred previous = u previous
          ptsdc=dc     ! u previous time step = u       
          DO i=1,mesh%nnodes
            DCtime(i)=-ptsdc(i)*inp%beta2+pptsdc(i)*inp%beta3
          END DO
        END IF


C      ----------------------------------------------------------------------------
C                                 START NONLINEAR LOOP
C      ----------------------------------------------------------------------------
2718    CONTINUE
          cnt%tsnnlit=cnt%tsnnlit+1  ! number of nonlinear iterations within time step
          cnt%glnnlit=cnt%glnnlit+1  ! number of all nonlinear iterations
C      ----------------------------------------------------------------------------
C                                 BOUNDARY VORTICITIES
C      ----------------------------------------------------------------------------

          IF (inp%iBw.EQ.1) THEN
          eqn=1
          CALL sdkm_solve(eqn,env,inp,cpu,mesh,cnt%nit(eqn),sdH,sdHtx,sdHty,sdHtz,sdDx,sdDy,sdDz,
     &                    sdAx,sdPivotx,velocity,vorticity)

          eqn=2
          CALL sdkm_solve(eqn,env,inp,cpu,mesh,cnt%nit(eqn),sdH,sdHtx,sdHty,sdHtz,sdDx,sdDy,sdDz,
     &                    sdAy,sdPivoty,velocity,vorticity)

          eqn=3
          CALL sdkm_solve(eqn,env,inp,cpu,mesh,cnt%nit(eqn),sdH,sdHtx,sdHty,sdHtz,sdDx,sdDy,sdDz,
     &                    sdAz,sdPivotz,velocity,vorticity)


          ELSE IF (inp%iBw.EQ.2) THEN
          eqn=1
          CALL sdkm_solve8(eqn,env,inp,cpu,mesh,cnt%nit(eqn),sdH,sdHtx,sdHty,sdHtz,sdDx,sdDy,sdDz,sdDxB,sdDyB,sdDzB,
     &                    sdAx,sdPivotx,velocity,vorticity)

          eqn=2
          CALL sdkm_solve8(eqn,env,inp,cpu,mesh,cnt%nit(eqn),sdH,sdHtx,sdHty,sdHtz,sdDx,sdDy,sdDz,sdDxB,sdDyB,sdDzB,
     &                    sdAy,sdPivoty,velocity,vorticity)

          eqn=3
          CALL sdkm_solve8(eqn,env,inp,cpu,mesh,cnt%nit(eqn),sdH,sdHtx,sdHty,sdHtz,sdDx,sdDy,sdDz,sdDxB,sdDyB,sdDzB,
     &                    sdAz,sdPivotz,velocity,vorticity)

          ELSE IF (inp%iBw.EQ.3) THEN

          eqn=1
          CALL sdkm_solveBCN(eqn,env,inp,cpu,mesh,cnt%nit(eqn),sdH,sdHtx,sdHty,sdHtz,sdDx,sdDy,sdDz,
     &                    sdAx,sdPivotx,velocity,vorticity)

          eqn=2
          CALL sdkm_solveBCN(eqn,env,inp,cpu,mesh,cnt%nit(eqn),sdH,sdHtx,sdHty,sdHtz,sdDx,sdDy,sdDz,
     &                    sdAy,sdPivoty,velocity,vorticity)

          eqn=3
          CALL sdkm_solveBCN(eqn,env,inp,cpu,mesh,cnt%nit(eqn),sdH,sdHtx,sdHty,sdHtz,sdDx,sdDy,sdDz,
     &                    sdAz,sdPivotz,velocity,vorticity)

          END IF

C      ----------------------------------------------------------------------------
C                                 DOMAIN VELOCITIES
C      ----------------------------------------------------------------------------
          IF (inp%iDv.EQ.1) THEN
            eqn=1
            CALL solv_kmMacro3(eqn,inp,cpu,mesh,cnt%nit(3+eqn),KMsysm,KMrhsm,KMprecv,kmp,
     &                       smatHtx,smatHty,smatHtz,
     &                       smatDx,smatDy,smatDz,velocity,vorticity)

            eqn=2
            CALL solv_kmMacro3(eqn,inp,cpu,mesh,cnt%nit(3+eqn),KMsysm,KMrhsm,KMprecv,kmp,
     &                       smatHtx,smatHty,smatHtz,
     &                       smatDx,smatDy,smatDz,velocity,vorticity)

            eqn=3
            CALL solv_kmMacro3(eqn,inp,cpu,mesh,cnt%nit(3+eqn),KMsysm,KMrhsm,KMprecv,kmp,
     &                       smatHtx,smatHty,smatHtz,
     &                       smatDx,smatDy,smatDz,velocity,vorticity)
          ELSE IF (inp%iDv.EQ.2) THEN
            eqn=1
            CALL solv_kmMacro3sq(eqn,inp,cpu,mesh,cnt%nit(3+eqn),KMsysm,KMrhsm,KMprec,kmp,
     &                       smatHtx,smatHty,smatHtz,
     &                       smatDx,smatDy,smatDz,velocity,vorticity)             

            eqn=2
            CALL solv_kmMacro3sq(eqn,inp,cpu,mesh,cnt%nit(3+eqn),KMsysm,KMrhsm,KMprec,kmp,
     &                       smatHtx,smatHty,smatHtz,
     &                       smatDx,smatDy,smatDz,velocity,vorticity)             

            eqn=3
            CALL solv_kmMacro3sq(eqn,inp,cpu,mesh,cnt%nit(3+eqn),KMsysm,KMrhsm,KMprec,kmp,
     &                       smatHtx,smatHty,smatHtz,
     &                       smatDx,smatDy,smatDz,velocity,vorticity)    
          END IF
C      ----------------------------------------------------------------------------
C                                   DOMAIN TEMPERATURES
C      ----------------------------------------------------------------------------
          IF (inp%iDT.EQ.1) THEN
            CALL SolveFDdae(env,io,inp,cpu,mesh,Tsysm,Trhsm,Tprecv,temp,qtemp,velocity,Ttime,
     &                      smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,
     &                      cnt%nit(10),inp%Ren*inp%prn/nanoC)
          ELSE IF (inp%iDT.EQ.2) THEN
            CALL SolveFDdaeSQ(env,io,inp,cpu,mesh,Tsysm,Trhsm,Tprec,temp,qtemp,velocity,Ttime,
     &                      smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,
     &                      cnt%nit(10),inp%Ren*inp%prn/nanoC)
          ELSE IF (inp%iDT.EQ.3) THEN
c           set up density, heat capacity, thermalconductivity and  Energy equation boundary conditions
            CALL SetUpRhoCpLambdaAndBC(env,io,inp,mesh,gauss,Density,HeatCapacity,
     &                                 ThermalConductivity,Temp,qTemp,velocity,inp%eeqm,Trhsv,cnt%rtime,0,PartVolFrac)
c           interpolate diffusivity to macro mesh
            CALL Int2MacroNodes(mesh,ThermalConductivity,qThermalConductivity)
C           Calculate gradient of thermal conductivity
            CALL setGrad(mesh,ThermalConductivity,gradTC)
c           system matrix
            CALL sMat2crsSysRhsB_TempRCPL_fill(mesh,smatH,smatG,smatB,inp%beta,inp%ren*inp%prn/nanoC,Tsysm,Trhsm,
     &                      smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,
     &                      Density,HeatCapacity,ThermalConductivity,qThermalConductivity,dRhoCpdt,velocity,gradTC)
c           preconditioner
            CALL prelsqr2(Tsysm%neq,mesh%Tnunk,Tsysm%nnz,Tprecv,Tsysm%i,Tsysm%j,Tsysm%v)
c           solve
            CALL SolveFDdaeRCPL(env,io,inp,cpu,mesh,Tsysm,Trhsm,Tprecv,temp,qtemp,velocity,Ttime,Trhsv,
     &                          smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,
     &                          cnt%nit(10),inp%Ren*inp%prn/nanoC,
     &                          Density,HeatCapacity,ThermalConductivity,gradTC)

          END IF
C      ----------------------------------------------------------------------------
C                                   KELVIN BODY FORCE
C      ----------------------------------------------------------------------------
          IF (inp%iFerro.GT.0) THEN
c           calculate temperature gradient
            CALL setGrad(mesh,temp,tempGrad)
            CALL CalKelvinBodyForce(mesh,inp,temp,mfs,hshg,kbf,tempGrad)
          END IF

C      ----------------------------------------------------------------------------
C                                   DOMAIN VORTICITIES
C      ----------------------------------------------------------------------------

          IF (inp%iDw.EQ.1) THEN
            eqn=1
            CALL SolveFDwTE(eqn,env,io,inp,cpu,mesh,VTEsysmX,VTErhsmX,VTEprecvX,
     &                    vorticity,qvorticity,temp,dc,velocity,kbf,rhsv(:,eqn),
     &                    smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,cnt%nit(6+eqn),nanoA,nanoB)

            eqn=2
            CALL SolveFDwTE(eqn,env,io,inp,cpu,mesh,VTEsysmY,VTErhsmY,VTEprecvY,
     &                    vorticity,qvorticity,temp,dc,velocity,kbf,rhsv(:,eqn),
     &                    smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,cnt%nit(6+eqn),nanoA,nanoB)

            eqn=3
            CALL SolveFDwTE(eqn,env,io,inp,cpu,mesh,VTEsysmZ,VTErhsmZ,VTEprecvZ,
     &                    vorticity,qvorticity,temp,dc,velocity,kbf,rhsv(:,eqn),
     &                    smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,cnt%nit(6+eqn),nanoA,nanoB)


          ELSE IF (inp%iDw.EQ.2) THEN
            eqn=1
            CALL SolveFDwTEsq(eqn,env,io,inp,cpu,mesh,VTEsysmX,VTErhsmX,VTEprecX,
     &                    vorticity,qvorticity,temp,dc,velocity,kbf,rhsv(:,eqn),
     &                    smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,cnt%nit(6+eqn),nanoA,nanoB)
            eqn=2
            CALL SolveFDwTEsq(eqn,env,io,inp,cpu,mesh,VTEsysmY,VTErhsmY,VTEprecY,
     &                    vorticity,qvorticity,temp,dc,velocity,kbf,rhsv(:,eqn),
     &                    smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,cnt%nit(6+eqn),nanoA,nanoB)
            eqn=3
            CALL SolveFDwTEsq(eqn,env,io,inp,cpu,mesh,VTEsysmZ,VTErhsmZ,VTEprecZ,
     &                    vorticity,qvorticity,temp,dc,velocity,kbf,rhsv(:,eqn),
     &                    smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,cnt%nit(6+eqn),nanoA,nanoB)

          ELSE IF (inp%iDw.EQ.3) THEN
c           use model to set up viscosity and cvte
            CALL SetUpViscCvte(env,io,inp,mesh,viscosity,cvte,temp)
c           interpolate Viscosity to macro mesh
            CALL Int2MacroNodes(mesh,Viscosity,qViscosity)
C           Calculate gradient of Viscosity
            CALL setGrad(mesh,Viscosity,GradVisc)
C           set up system matrices

            CALL sMat2crsSysRhsB_w_fill_vMU(1,mesh,smatH,smatG,smatB,inp%beta,inp%ren/nanoA,VTEsysmX,VTErhsmX,
     &                                velocity,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,Viscosity,qViscosity)
            CALL sMat2crsSysRhsB_w_fill_vMU(2,mesh,smatH,smatG,smatB,inp%beta,inp%ren/nanoA,VTEsysmY,VTErhsmY,
     &                                velocity,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,Viscosity,qViscosity)
            CALL sMat2crsSysRhsB_w_fill_vMU(3,mesh,smatH,smatG,smatB,inp%beta,inp%ren/nanoA,VTEsysmZ,VTErhsmZ,
     &                                velocity,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,Viscosity,qViscosity)

c           calculate preconditioner
            CALL prelsqr2(VTEsysmX%neq,mesh%nunk(1),VTEsysmX%nnz,VTEprecvX,VTEsysmX%i,VTEsysmX%j,VTEsysmX%v)
            CALL prelsqr2(VTEsysmY%neq,mesh%nunk(2),VTEsysmY%nnz,VTEprecvY,VTEsysmY%i,VTEsysmY%j,VTEsysmY%v)
            CALL prelsqr2(VTEsysmZ%neq,mesh%nunk(3),VTEsysmZ%nnz,VTEprecvZ,VTEsysmZ%i,VTEsysmZ%j,VTEsysmZ%v)
c           calculate terms that involve viscosity gradient
            CALL ViscGradTerms(mesh,velocity,vorticity,GradVisc,DvaEpsGVisc,CuOmCrGrVi)
C
C           solve
C
            eqn=1
            CALL SolveFDwTE_vMU(eqn,env,io,inp,cpu,mesh,VTEsysmX,VTErhsmX,VTEprecvX,
     &                    vorticity,qvorticity,temp,dc,velocity,kbf,rhsv(:,eqn),
     &                    smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,cnt%nit(6+eqn),nanoA,nanoB,
     &                    DvaEpsGVisc,CuOmCrGrVi,cvte)

            eqn=2
            CALL SolveFDwTE_vMU(eqn,env,io,inp,cpu,mesh,VTEsysmY,VTErhsmY,VTEprecvY,
     &                    vorticity,qvorticity,temp,dc,velocity,kbf,rhsv(:,eqn),
     &                    smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,cnt%nit(6+eqn),nanoA,nanoB,
     &                    DvaEpsGVisc,CuOmCrGrVi,cvte)

            eqn=3
            CALL SolveFDwTE_vMU(eqn,env,io,inp,cpu,mesh,VTEsysmZ,VTErhsmZ,VTEprecvZ,
     &                    vorticity,qvorticity,temp,dc,velocity,kbf,rhsv(:,eqn),
     &                    smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,cnt%nit(6+eqn),nanoA,nanoB,
     &                    DvaEpsGVisc,CuOmCrGrVi,cvte)


          END IF 

C      ----------------------------------------------------------------------------
C                                   GENERAL DIFF-CONV EQUATION
C      ----------------------------------------------------------------------------
          IF (inp%iDC.GT.0) THEN
c           set up diffusivity and sources
            CALL SetUpDiffSources(mesh,inp,DCdiff,DCrhsv,DCtime)

            IF (inp%iDC.EQ.1) THEN  ! verzija, ki rabi laplace
c             calculate difusivitiy gradients
              CALL CalDiffGrad(mesh,DCdiff,DCdiffGrad,DCdiffLap)
c             set up system matrix
              CALL sMat2crsSysRhsB_DC_fill(mesh,smatH,smatG,smatB,inp%beta,DCdiff,DCsysm,DCrhsm,
     &                               velocity,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz)              
c             preconditioner
              CALL prelsqr2(Dcsysm%neq,mesh%DCnunk,DCsysm%nnz,Dcprecv,DCsysm%i,Dcsysm%j,DCsysm%v)
c             solve
              CALL SolveFDdaeVD(env,io,inp,cpu,mesh,DCsysm,DCrhsm,DCprecv,dc,qdc,velocity,DCrhsv,
     &                      smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,
     &                      cnt%nit(11),DCdiff,DCdiffGrad,DCdiffLap)
            ELSE IF (inp%iDC.EQ.2) THEN ! verzija ki ne rabi Laplace
c             calculate difusivitiy gradients
              CALL setGrad(mesh,DCdiff,DCdiffGrad)
c             interpolate diffusivity to macro mesh
              CALL Int2MacroNodes(mesh,DCdiff,DCqDiff)
c             set up system matrix
              CALL sMat2crsSysRhsB_DC_fill2(mesh,smatH,smatG,smatB,inp%beta,DCdiff,DCqDiff,DCsysm,DCrhsm,
     &                               velocity,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz)
c             preconditioner
              CALL prelsqr2(Dcsysm%neq,mesh%DCnunk,DCsysm%nnz,Dcprecv,DCsysm%i,Dcsysm%j,DCsysm%v)
c             solve
              CALL SolveFDdaeVD2(env,io,inp,cpu,mesh,DCsysm,DCrhsm,DCprecv,dc,qdc,velocity,DCrhsv,
     &                      smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,
     &                      cnt%nit(11),DCdiff,DCdiffGrad)

            ELSE IF (inp%iDC.EQ.3) THEN ! verzija ki ne rabi Laplace, DOLOCEN SISTEM ENACB

c             calculate difusivitiy gradients
              CALL setGrad(mesh,DCdiff,DCdiffGrad)
c             interpolate diffusivity to macro mesh
              CALL Int2MacroNodes(mesh,DCdiff,DCqDiff)
c             set up system matrix
              CALL sMat2crsSysRhsB_DC_fill2sq(mesh,smatH,smatG,smatB,inp%beta,DCdiff,DCqDiff,DCsysm,DCrhsm,
     &                               velocity,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz)
c             preconditioner
              CALL dcopy (DCprec%nnz, DCsysm%v, 1, DCprec%v,1)
              CALL FormPRMjr(inp%sqrs_prec,DCprec,DCsysm)
c             solve
              CALL SolveFDdaeVD2sq(env,io,inp,cpu,mesh,DCsysm,DCrhsm,DCprec,dc,qdc,velocity,DCrhsv,
     &                      smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,
     &                      cnt%nit(11),DCdiff,DCdiffGrad)

            ELSE IF (inp%iDC.EQ.4) THEN ! verzija ki ne rabi Laplace, IN JE ZA LINEARNE PROBLEME
c             calculate difusivitiy gradients
              CALL setGrad(mesh,DCdiff,DCdiffGrad)
c             interpolate diffusivity to macro mesh
              CALL Int2MacroNodes(mesh,DCdiff,DCqDiff)
c             set up system matrix
              CALL sMat2crsSysRhsB_DC_fill4(mesh,smatH,smatG,smatB,inp%beta,DCdiff,DCdiffGrad,DCqDiff,DCsysm,DCrhsm,
     &                               velocity,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz)
c             preconditioner
              CALL prelsqr2(Dcsysm%neq,mesh%DCnunk,DCsysm%nnz,Dcprecv,DCsysm%i,Dcsysm%j,DCsysm%v)
c             solve
              CALL SolveFDdaeVD4(env,io,inp,cpu,mesh,DCsysm,DCrhsm,DCprecv,dc,qdc,velocity,DCrhsv,
     &                      smatB,smatAbdx,smatAbdy,smatAbdz,smatDx,smatDy,smatDz,
     &                      cnt%nit(11),DCdiff,DCdiffGrad)



            END IF
          END IF

C      ----------------------------------------------------------------------------
C                                   CHECK CONVERGENCE
C      ----------------------------------------------------------------------------
          
          CALL CheckConvergencevw(env,io,inp,mesh,cnt,velocity,pnlsvelocity,vorticity,pnlsvorticity,
     &                            temp,pnlstemp,dc,pnlsdc)
     
          CALL CheckControlFiles(env,io,inp,istop) 
          
          IF (inp%iDLSE.EQ.1.AND.cnt%tsnnlit.GT.1) THEN  ! prva dva naredi s 1E-7, nato dvigne
            IF (cnt%tsnnlit.EQ.2) inp%dlse=inp%dlse_epsth
            CALL DLSE(env,io,cnt,inp)
          END IF
C      ----------------------------------------------------------------------------
C                                 END NONLINEAR LOOP
C      ----------------------------------------------------------------------------
        IF (cnt%tsnnlit.LT.inp%nnlit
     &                                .AND.(cnt%nlierr(1).GT.inp%eps
     &                                  .OR.cnt%nlierr(2).GT.inp%eps
     &                                  .OR.cnt%nlierr(3).GT.inp%eps
     &                                  .OR.cnt%nlierr(4).GT.inp%eps
     &                                  .OR.cnt%nlierr(5).GT.inp%eps
     &                                  .OR.cnt%nlierr(6).GT.inp%eps
     &                                  .OR.cnt%nlierr(7).GT.inp%eps     
     &                                  .OR.cnt%nlierr(8).GT.inp%eps)     
     &                                .AND.istop.EQ.0) GOTO 2718

C      ----------------------------------------------------------------------------
C                                 MOVE PARTICLES
C      ----------------------------------------------------------------------------
      IF (inp%iPart.GT.0) THEN
        CALL MoveParticles(inp,part,mesh,cpu,velocity,vorticity,Temp)
        CALL CalParticleVolumeFraction(env,io,inp,gauss,part,mesh,PartVolFrac)
      END IF

        IF (env%myproc.EQ.1) THEN
          IF ((MOD(cnt%trtstep,inp%iwrit).EQ.0).OR.istop.NE.0) THEN
c           izpise tri.walls.dat
            IF (inp%iRawl.GT.0) CALL ExportWallToTecplot(io,inp,cnt,mesh,velocity,vorticity,temp,dc,qtemp,qdc)
C           tri.rms
            IF (inp%eeqm.GT.10.AND.inp%iDT.EQ.3) CALL CalTempRMS(env,io,inp,mesh,cnt,Temp,cnt%rtime,inp%eeqm)
c           izpise delce
            IF (inp%iPart.GT.0) CALL OutputParticles(io,inp,cnt,part)
c           izpise tri.materials.dat
            IF (inp%iDt.EQ.3) CALL OutputTDMaterialProperties(io,inp,cnt,mesh,Density,HeatCapacity,ThermalConductivity,PartVolFrac)
c           izpise tri.dat
            CALL OutputTDResultsTecplotvw(io,inp,cnt,mesh,velocity,vorticity,temp,dc)
C           izpise rezultate po stenah (tri.twfl.dat)
            CALL ExportWallFlux(io%twfl,mesh,cnt,qtemp,qvorticity)
c           izpise tri.vtu
            IF (inp%iPara.GT.0) CALL OutputTDResultsParaview(io,inp,cnt,mesh,velocity,vorticity,temp,dc)
          END IF
        END IF

        CALL CheckTIMEConvergence(env,io,mesh,cnt,vorticity,ptsvorticity)
        
        IF (env%myproc.EQ.1) THEN
          IF (inp%iDT.EQ.1.OR.inp%iDT.EQ.2) THEN
            CALL WallFlux(io%tfl,mesh,cnt,gauss,qtemp," qT")
            CALL WallFuncInt(io%tfl,mesh,cnt,gauss,temp,"  T")
            CALL WallLocalFlux(io%lnu,mesh,cnt,gauss,qtemp)
          ELSE IF (inp%iDT.EQ.3) THEN
            CALL WallFlux(io%tfl,mesh,cnt,gauss,qtemp," qT")
            CALL WallFluxLambda(io%tfl,mesh,cnt,gauss,qtemp,"qTl",ThermalConductivity)
          END IF

          CALL MassFlux(io%mass,mesh,cnt,gauss,velocity)

          CALL WallFlux(io%wfl,mesh,cnt,gauss,qvorticity(:,1),"qwx")
          CALL WallFlux(io%wfl,mesh,cnt,gauss,qvorticity(:,2),"qwy")
          CALL WallFlux(io%wfl,mesh,cnt,gauss,qvorticity(:,3),"qwz")

          CALL WallFuncInt(io%wfl,mesh,cnt,gauss,vorticity(:,1)," wx")
          CALL WallFuncInt(io%wfl,mesh,cnt,gauss,vorticity(:,2)," wy")
          CALL WallFuncInt(io%wfl,mesh,cnt,gauss,vorticity(:,3)," wz")

          tmp=abs(vorticity(:,1))
          CALL WallFuncInt(io%wfl,mesh,cnt,gauss,tmp,"Awx")
          tmp=abs(vorticity(:,2))
          CALL WallFuncInt(io%wfl,mesh,cnt,gauss,tmp,"Awy")
          tmp=abs(vorticity(:,3))
          CALL WallFuncInt(io%wfl,mesh,cnt,gauss,tmp,"Awz")

          IF (inp%iDC.GT.0) THEN
            CALL WallFlux(io%wfl,mesh,cnt,gauss,qdc,"qdc") ! zapisem kar v worticity flux file
          END IF

C         Write restart file
          CALL WrstF(mesh,io,inp,cnt,velocity,vorticity,
     &             ptsvorticity,pptsvorticity,qvorticity,pnlsvorticity,pnlsvelocity,
     &             temp,ptstemp,pptstemp,qtemp,pnlstemp,
     &             dc,ptsdc,pptsdc,qdc,pnlsdc)
         END IF

C      ----------------------------------------------------------------------------
C                                 END TIME LOOP
C      ----------------------------------------------------------------------------
      IF (cnt%trtstep.LT.inp%nstep.AND.istop.EQ.0) GOTO 1618               

      CALL VmesniCas(cpu,itim)
      cpu%time(itim+1)=cptime(cpu%t00)        
c
C____ Write cpu information
c
      IF (env%myproc.EQ.1) CALL WriteCpu(io,cpu)        
c
C____ Finish
c      
      CALL StopProgram(env,io,inp,0)    
      END
